#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_tables.h>
#include <cyg/hal/plf_stub.h>
#include <cyg/infra/diag.h>
#include <network.h>
#include "../dm.h"
#include <cyg/hal/iic.h>
#include "../ident.h"

/* tadmin task thread variables */
static unsigned tadminStack[1024];
static cyg_handle_t  tadminHandle;
static cyg_thread    tadminThread;
static cyg_thread_entry_t tadminTask;

/*
 *  Init function for the MAC mode admin thread.  This task implements Comtrol
 *  admin functions, that allow device control, administration and troubleshooting.
 */
void tadminInit(void)
{
  diag_printf("starting tadmin\n");
  cyg_thread_create(30, tadminTask, (cyg_addrword_t)0, "tadmin",
                    (void *)tadminStack, sizeof tadminStack, 
                    &tadminHandle, &tadminThread);
  cyg_thread_resume(tadminHandle);
}

typedef struct
{
  unsigned char	 cmd;
  unsigned char	 index;
  unsigned short length;
  unsigned long	 addr;
} __attribute__((packed)) tAdminReq;

typedef struct 
{
  unsigned long   IpAddr;
  unsigned long   DefaultRoute;
  unsigned long   ServerIpAddr;
  unsigned long   SubnetMask;
  unsigned long   AvailMemAddr;
  unsigned long   AvailMemLength;
  char  Version[64];
  unsigned char   MacAddr[6];
  unsigned long   ModelId;
  unsigned char   ArchId;
  unsigned char   NumPorts;
  unsigned short   ProgramId;
  unsigned char   BoardRev;
  unsigned char   BootRevMajor;
  unsigned char   BootRevMinor;
  unsigned char   NumApps;
  struct 
    {
      unsigned short AppId;
      unsigned char AppRevMajor;
      unsigned char AppRevMinor;
    }AppData __attribute__((packed));
  unsigned char   UserDataCount;
} __attribute__((packed)) tIdResponse;

#define LOAD_TOOLREQ	1
#define GET_TOOLREQ	2
#define GOTO_TOOLREQ	3
#define MOVE_TOOLREQ	4
#define ID_TOOLREQ	5
#define SET_TOOLREQ	6
#define ENV_TOOLREQ	7
#define PING_TOOLREQ	8
#define DATA_TOOLREQ	9
#define SET_IP_TOOLREQ  10
#define ERASE_IP_TOOLREQ 11
#define DISABLE_IP_TOOLREQ 12
#define RESET_TOOLREQ   13

static void handleAdminCommand(int sock, tAdminReq *req, int len);
static unsigned char buf[1500];
static int resetRequest;

static void tadminTask(cyg_addrword_t data)
{
  int lsock;
  int one,rc;
  struct sockaddr_in slocal,sremote;
  extern int TcpStackOk;
  
  diag_printf("tadmin running\n");

  while (!TcpStackOk)
    cyg_thread_delay(100);
  
  lsock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (lsock < 0) 
    {
      diag_printf("Cannot create socket.\n");
      return;
    }
  
  memset(&slocal, 0, sizeof(slocal));
  slocal.sin_family = AF_INET;
  slocal.sin_len = sizeof(slocal);
  slocal.sin_port = 0x11fe;
  slocal.sin_addr.s_addr = INADDR_ANY;
  
  rc = bind(lsock, (struct sockaddr *)&slocal, sizeof(slocal));
  if (rc) 
    {
      diag_printf("Cannot bind socket: %d %d\n",rc,errno);
    }
  
  one = 1;
  if (setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one))) 
    diag_printf("Cannot set ReuseAddr\n");
  if (setsockopt(lsock, SOL_SOCKET, SO_REUSEPORT, &one, sizeof(one))) 
    diag_printf("Cannot set ReusePort\n");
  
  listen(lsock, 1);
  
  while (1)
    {
      int dsock;
      int len = sizeof(sremote);
      dsock = accept(lsock, (struct sockaddr *)&sremote, &len);
      if (dsock < 0)
        diag_printf("Failure on accept\n");
      diag_printf("tadmin: connection from %s:%d\n",
                  inet_ntoa(sremote.sin_addr),sremote.sin_port);
      while (1)
        {
          len = read(dsock,buf,sizeof buf);
          if (len <= 0)
            break;
          handleAdminCommand(dsock,(tAdminReq*)buf,len);
        }
      close(dsock);
      if (resetRequest)
        {
          cyg_thread_delay(1);
          HAL_STUB_PLATFORM_RESET();
        }
    }
}

static void handleAdminCommand(int sock, tAdminReq *req, int len)
{
  extern char __bss_end[];
  
  if (len < sizeof *req)
    {
      diag_printf("Tcp admin request too short: %d\n",len);
      return;
    }
  
  switch (req->cmd)
    {
     case RESET_TOOLREQ:
      diag_printf("Tcp: reset\n");
      resetRequest = 1;
      break;
      
     case ERASE_IP_TOOLREQ:
      diag_printf("Tcp: erase IP info\n");
      putIpInfoRt(0,0,0);
      break;
      
     case DISABLE_IP_TOOLREQ:
      diag_printf("Tcp: erase IP info\n");
      putIpInfoRt(0xffffffff,0,0);
      break;
      
     case SET_IP_TOOLREQ:
        {
          unsigned long *p = ((unsigned long*)(&req->addr)) + 1;
          diag_printf("Tcp: set IP %08x %08x %08x\n",p[0],p[1],p[2]);
          putIpInfoRt(p[0],p[1],p[2]);
          break;
        }
      
     case ID_TOOLREQ:
        {
          tIdResponse tresp;
          tresp.IpAddr = ident.ipAddr;
          tresp.DefaultRoute = ident.gateway;
          tresp.ServerIpAddr = 0;
          tresp.SubnetMask = ident.netMask;
          tresp.AvailMemAddr = swap32(htonl((int)__bss_end));
          tresp.AvailMemLength = swap32(htonl((int)(((char*)0x7c0000) - __bss_end)));
          memcpy(tresp.Version, versionString, 64);
          memcpy(tresp.MacAddr, ident.macAddr, 6);
          tresp.ModelId = htonl(ident.modelId);
          tresp.ArchId = ident.archId;
          tresp.NumPorts = ident.numPorts;
          tresp.ProgramId = htons(2);
          tresp.BoardRev = ident.boardRev;
          tresp.BootRevMajor = 0;
          tresp.BootRevMinor = 0;
          tresp.NumApps = 1;
          tresp.AppData.AppId = htons(2);
          tresp.AppData.AppRevMajor = ident.versionMajor;
          tresp.AppData.AppRevMinor = ident.versionMinor;
          tresp.UserDataCount = 0;
          write(sock,&tresp, sizeof tresp);
          break;
        }
      
     default:
      diag_printf("Tcp: unrecognized command: %d\n",req->cmd);
      break;
    }
}
