#include <pkgconf/kernel.h>
#include <pkgconf/hal.h>                // CYG_HAL_USE_ROM_MONITOR_CYGMON
#include <cyg/kernel/ktypes.h>
#include <cyg/kernel/thread.hxx>
#include <cyg/kernel/sched.hxx>
#include <cyg/kernel/thread.inl>
#include <cyg/kernel/sched.inl>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_stub.h>
#include <cyg/infra/diag.h>
#include <cyg/kernel/kapi.h>

#include "watchdog.h"

static void dumpThreadInfo(Cyg_Thread *thread)
{
static char *sleepStrings[] = {"none","wait","delay","timeout","break","destruct","exit","done"};
  extern cyg_uint32 __etext,__stext,__ram_data_start,__ram_data_end,__bss_start,__bss_end;
  cyg_uint32 *sp, last;
  cyg_uint32 stackBase, stackSize;
  int repcount;
  unsigned state, wakeReason, sleepReason;
  
  state = thread->get_state();
  stackBase = thread->get_stack_base();
  stackSize = thread->get_stack_size();
  sleepReason = thread->get_sleep_reason();
  wakeReason = thread->get_wake_reason();
  
  if( thread == 0 ) 
    return;

#ifdef CYGVAR_KERNEL_THREADS_NAME
    diag_printf("%s: ",thread->get_name());
#else
    diag_printf("??: ");
#endif

  cyg_thread_delay(2);
  diag_printf("pri=%2d state=", thread->get_priority());
  cyg_thread_delay(2);
  
  if (state & Cyg_Thread::SUSPENDED)
    diag_printf("susp+");
  switch( state & ~Cyg_Thread::SUSPENDED )
    {
     case Cyg_Thread::RUNNING:
        if ( Cyg_Scheduler::get_current_thread() == thread ) 
        {
          diag_printf("running ");
          break;
        }
      else if ( state & Cyg_Thread::SUSPENDED ) 
          break;
      else 
        {
          diag_printf("ready   ");
          break;
        }
    case Cyg_Thread::SLEEPING:
      diag_printf("sleep[%s] %d",sleepStrings[sleepReason]);
      break;
    case Cyg_Thread::COUNTSLEEP | Cyg_Thread::SLEEPING:
    case Cyg_Thread::COUNTSLEEP:
        diag_printf("cntdsleep");
      break;
     case Cyg_Thread::CREATING:
      diag_printf("creating");
      break;
    case Cyg_Thread::EXITED:
      diag_printf("exited  ");
      break;
    default:
      diag_printf("unknown ");
      break;
    }

  cyg_thread_delay(2);
  diag_printf(" stack=%06x:%05x\n",stackBase,stackSize);
  
  HAL_SavedRegisters *regs = thread->get_saved_context();
  if (regs)
    {
      // dump reg values
#     define dumpreg(r) {diag_printf(" "#r "=%08x\n",regs->r);cyg_thread_delay(2);}
      dumpreg(fp);
      dumpreg(sp);
      dumpreg(lr);
      dumpreg(pc);
      dumpreg(cpsr);
      dumpreg(ip);
    }
  
  if (thread == Cyg_Scheduler::get_current_thread())
    return;
  
  diag_printf(" stack:\n");
  last = 0xdeadbeef;
  repcount = 0;
  
  if (regs->fp > stackBase && regs->fp < stackBase+stackSize)
    sp = (cyg_uint32*)regs->fp;
  else if (regs->sp > stackBase && regs->sp < stackBase+stackSize)
    sp = (cyg_uint32*)regs->sp;
  else 
    sp = (cyg_uint32*)stackBase;
  
  for (; sp < (cyg_uint32*)(stackBase+stackSize); ++sp)
    {
      watchdogReset();
      if (*sp == last)
        {
          ++repcount;
          continue;
        }
      if (repcount)
        diag_printf(" [repeats %d times]\n",repcount+1);
      else if (last != 0xdeadbeef)
        diag_printf("\n");
      cyg_thread_delay(2);
      repcount = 0;
      last = *sp;
      diag_printf("  %08x",*sp);
      if (*sp >= (cyg_uint32)&__stext && *sp <= (cyg_uint32)&__etext)
        diag_printf(" T");
      else if (*sp >= (cyg_uint32)&__etext && *sp <= (cyg_uint32)&__ram_data_start)
        diag_printf(" R");
      else if (*sp >= (cyg_uint32)&__ram_data_start && *sp <= (cyg_uint32)&__ram_data_end)
        diag_printf(" D");
      else if (*sp >= stackBase && *sp <= stackBase+stackSize)
        diag_printf(" S");
      else if (*sp >= (cyg_uint32)&__bss_start && *sp <= (cyg_uint32)&__bss_end)
        diag_printf(" B");
    }
  diag_printf("\n");
}


externC void dumpThreads(void)
{
  Cyg_Thread *th = Cyg_Thread::get_list_head();
  while( th != 0 )
    {
      dumpThreadInfo(th);
      cyg_thread_delay(5);
      th = th->get_list_next();
    }
}


