/*******************************************************************
 *  portmon.c
 *
 *  Implements the Debug Port Monitor function.  This allows the user
 *  to connect to the socket at port 4607 and receive the text that
 *  normally goes to out the debug port (UART 1).
 *******************************************************************/
#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_tables.h>
#include <cyg/hal/plf_stub.h>
#include <cyg/infra/diag.h>
#include <network.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "dm.h"

/*  Statics Variables */
static unsigned portmonStack[2048];
static cyg_handle_t  portmonHandle;
static cyg_thread    portmonThread;
static cyg_thread_entry_t portmonTask;
static int dsock = 0, portmonConnectionActive;

/* 
 *  Init function for the portmon task.  This thread allows you to attack via 
 *  telnet at port 4607 into the device and receive the debug header text.
 */
void portmonInit(void)
{
//    diag_printf("starting Debug Port Monitor \n");
    cyg_thread_create(30, 
		      portmonTask, 
		      (cyg_addrword_t)0, 
		      "portmon",
		      (void *)portmonStack, 
		      sizeof portmonStack, 
		      &portmonHandle, 
		      &portmonThread);
    cyg_thread_resume(portmonHandle);
}

/*  An eCos thread that listens on a socket at port 4607. */
static void portmonTask(cyg_addrword_t data)
{
    int lsock, len, rc;
    struct sockaddr_in slocal,sremote;
    extern int TcpStackOk;
    
//    diag_printf("Debug Port Monitor running\n");
    
    while (!TcpStackOk)
	cyg_thread_delay(50);
    
    lsock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (lsock < 0) {
	diag_printf("Debug Port Monitor:Cannot create socket.\n");
	return;
    }
    
    memset(&slocal, 0, sizeof(slocal));
    slocal.sin_family = AF_INET;
    slocal.sin_len = sizeof(slocal);
    slocal.sin_port = 4607; 
    slocal.sin_addr.s_addr = INADDR_ANY;
 
    rc = bind(lsock, (struct sockaddr *)&slocal, sizeof(slocal));
    if (rc) {
      diag_printf("Debug Port Monitor:Cannot bind socket: %s\n", strerror(errno));
      return;
    }
    listen(lsock, 1);
  
    while (1)  {
	portmonConnectionActive = 0;
	len = sizeof(sremote);
	dsock = accept(lsock, (struct sockaddr *)&sremote, &len);
	if (dsock < 0)  
	    diag_printf("Debug Port Monitor: failure on accept\n");
	diag_printf("Debug Port Monitor: connection from %s:%d\n", inet_ntoa(sremote.sin_addr),sremote.sin_port);
	portmonConnectionActive = 1;
    }
}

/*  Function that interrupts diag_printf.  If the Debug Port Monitor
    socket is open, the text is also sent there. */
void ddiag_printf(const char *fmt, ...)
{
    static char str[80];
    char work_str[80];
    CYG_ADDRWORD args[8];
    int rc;

    /*  There is no way to send varargs into another varargs function.
        Must gather the args and send them manually. */
    va_list a;
    va_start(a, fmt);

    /* Move all of the arguments into simple scalars. This avoids
     * having to use varargs to define diag_vprintf().   */
    args[0] = va_arg(a,CYG_ADDRWORD);
    args[1] = va_arg(a,CYG_ADDRWORD);
    args[2] = va_arg(a,CYG_ADDRWORD);
    args[3] = va_arg(a,CYG_ADDRWORD);
    args[4] = va_arg(a,CYG_ADDRWORD);
    args[5] = va_arg(a,CYG_ADDRWORD);
    args[6] = va_arg(a,CYG_ADDRWORD);
    args[7] = va_arg(a,CYG_ADDRWORD);

    va_end(a);
    diag_vprintf(fmt, args);

    /*  If socket is open, send text there also. */
    if (dsock) {

	/*  Gather up text until a full line is received. */
	diag_vsprintf(work_str, fmt, args);
	strcat(str, work_str);
	if (str[strlen(str)-1] == 0x0A) {
	    strcat(str, "\r");   
	    rc = write(dsock, str, strlen(str));
	    if (rc == -1) {
		close(dsock);
		dsock = 0;
		ddiag_printf("Debug Port Monitor: connection closed \n");
	    }
	    str[0] = 0;
	}
    }
} 

int portmon_read_char_nonblocking(char *dst)
{
  int avail = 0;
  if (!dsock)
    return 0;
  if (ioctl(dsock,FIONREAD,&avail))
    return 0;
  if (!avail)
    return 0;
  if (read(dsock,dst,1) == 1)
    return 1;
  return 0;
}
