/**********************************************************************
 *
 *      state.h
 *
 *      Include file that contains structs that hold device and port status.
 *      Note the global array of structs declared at the bottom.
 *
 **********************************************************************/

#ifndef _STATE_H
#define _STATE_H

#if !defined(__ECOS)
typedef unsigned cyg_io_handle_t; 
typedef unsigned cyg_handle_t;
typedef unsigned cyg_thread;
#else
#include <cyg/kernel/kapi.h>
#include <cyg/io/io.h>
#endif


// Increment ConfigVersion any time the config structure
// changes in an incompatible way.
// 
// Adding things to the end of the struct is OK (doesn't cause
// compat issues) as long as you want a default value of 0 for
// the object.

#define ConfigVersion 0x12345607

typedef struct
{
  /* connect/disconnect configuration */
  int             enabled;
  int             listenEnable;
  unsigned short  listenPort;
  
  unsigned        connectToIP;
  unsigned short  connectToPort;

  int             connectAlways;
  int             connectOnData;
  int             connectOnCD;
  int             connectOnDSR;
  
  int             disconnOnIdle;
  int             disconnOnNoCD;
  int             disconnOnNoDSR;
  
  unsigned        idleTimeout;
  
  unsigned        icTimeout;
  
  /* serial port configuration */
  enum {Mode232, Mode422, Mode485} interfaceMode;
  enum {FlowNone, FlowRTSCTS, FlowXonXoff, FlowHalfDuplex} flowControl;
  enum {ParityNone, ParityEven, ParityOdd, ParityMark, ParitySpace} parity;
  enum {DTRoff, DTRon, DTRconnected} dtrControl;
  unsigned baudRate;
  unsigned wordLength;
  int stopBits;

  int telnetEnable;
  
  /* add new fields at bottom and assume they'll be 0 by default */
}tPortConfig;

typedef struct
  {
    tPortConfig config;
    
    cyg_mutex_t serverMutex;            // mutex used for intra-server threads
    cyg_sem_t socketSemaphore;
    int       lastListenPort;
    unsigned  idleTimer;
    int       configUpdated;
    
    /* general port status */
    enum {ServerIdle, ServerBusy, ServerClosing} state; // Connection active?

    int             socket;              // TCP socket to read/write
    int             serverNumber;        // serial port number 0-N
    unsigned long   locIP;               // local IP address
    unsigned short  locPort;             // local Port
    unsigned long   remIP;               // remote IP address
    unsigned short  remPort;             // remote Port
    cyg_sem_t       *ownership;          // semaphore for port ownership
    cyg_io_handle_t serialHandle;        // io handle for serial port
    unsigned long   rxCount;             // number of bytes received from TCP
    unsigned long   txCount;             // number of bytes sent to TCP
    
    /* state for task that reads from TCP for this port */
    unsigned        rxStack[2048];
    cyg_handle_t    rxHandle;
    cyg_thread      rxThread;
    
    /* state for task that writes to TCP for this port */
    unsigned        txStack[2048];
    cyg_handle_t    txHandle;
    cyg_thread      txThread;

  } tServerState;


/* global array of structures, one per serial port */
tServerState serverState[DevMastMaxPorts];


#endif /*  _STATE_H  */
