#include  "uemf.h"
#include  "websda.h"
#include  "wsIntrn.h"
#include  <network.h>
#include  "../ident.h"
  
/*  GoAhead uses a trace macro for debug printing.  Setting this
    define to 1 gets minimal, 99 gets all, 0 gets none. */
#define GOAHEAD_TRACE_LEVEL 1

static int finished;
static int initWebs();
static int websHomePageHandler(webs_t wp, char_t * urlPrefix, char_t * webDir,
			       int arg, char_t * url, char_t * path,
			       char_t * query);
extern void defaultErrorHandler(int etype, char_t * msg);
extern void defaultTraceHandler(int level, char_t * buf);
extern void deviceMasterSetAuthentication(unsigned int);

#if B_STATS
#error WARNING:  B_STATS directive is not supported in this OS!
#endif

#define loop()    do cyg_thread_delay(100); while (1)

//======================================================================
static void webServerTask(cyg_addrword_t data)
{
/*
 *	Initialize the memory allocator. Allow use of malloc and start 
 *	with a 60K heap.  For each page request approx 8KB is allocated.
 *	60KB allows for several concurrent page requests.  If more space
 *	is required, malloc will be used for the overflow.
 */
  bopen(NULL, (60 * 1024), B_USE_MALLOC);

/*
 *	Initialize the web server
 */
  if (initWebs() < 0)
    {
      diag_printf("Web server init failed\n");
      loop();
    }

/*
 *	Basic event loop. SocketReady returns true when a socket is ready for
 *	service. SocketSelect will block until an event occurs. SocketProcess
 *	will actually do the servicing.
 */
  while (!finished)
    {
      if (socketReady(-1) || socketSelect(-1, 2000))
        socketProcess(-1);
      emfSchedProcess();
    }

/*
 *	Close the socket module, report memory leaks and close the memory allocator
 */
  websCloseServer();
  socketClose();
  bclose();
  diag_printf("Web server closed\n");
  loop();
}

static int myWebsSecurityHandler(webs_t wp, char_t * urlPrefix, char_t * webDir, int arg, char_t * url, char_t * path, char_t * query);
extern void initLocalSupport(void);
extern int TcpStackOk;

//======================================================================
static int initWebs()
{
  char host[128];
  char_t wbuf[128];

  while (!TcpStackOk)
    cyg_thread_delay(10);

  socketOpen();
  
  websSetDefaultDir("/");
  ascToUni(wbuf, inet_ntoa(eth0_bootp_data.bp_yiaddr), sizeof(wbuf));
  websSetIpaddr(wbuf);
  ascToUni(wbuf, host, sizeof(wbuf));
  websSetHost(wbuf);
  websSetDefaultPage(T("default.asp"));
  websSetPassword(ident.password);
  if (ident.authMethod > 4)
    ident.authMethod = 0;
  deviceMasterSetAuthentication(ident.authMethod);

  websOpenServer(80, 0);
  websUrlHandlerDefine(T(""), NULL, 0, myWebsSecurityHandler, WEBS_HANDLER_FIRST);
  websUrlHandlerDefine(T("/goform"), NULL, 0, websFormHandler, 0);
  websUrlHandlerDefine(T(""), NULL, 0, websDefaultHandler, WEBS_HANDLER_LAST);
  websUrlHandlerDefine(T("/"), NULL, 0, websHomePageHandler, 0);

  initLocalSupport();
  
  return 0;
}

#define AuthNoaccess 0
#define AuthNone 1
#define AuthBasic 2
#define AuthMD5 3
#define AuthInvalid 4

static char passwordRequiredString[] = "Password Required";

//======================================================================
int myWebsSecurityHandler(webs_t wp, char_t * urlPrefix, char_t * webDir, int arg, char_t * url, char_t * path, char_t * query)
{
  char_t *password;
  int flags;
  
  a_assert(websValid(wp));
  a_assert(url && *url);
  a_assert(path && *path);
  
  password = websGetRequestPassword(wp);
  flags = websGetRequestFlags(wp);
  
  switch (ident.authMethod)
    {
     default:
     case AuthNone:
      goto loginOK;
      
     case AuthNoaccess:
      websError(wp, 400, "Page Not Found");
      goto loginFail;
      
     case AuthInvalid:
      websError(wp, 403, "Access Denied");
      goto loginFail;
      
     case AuthMD5:
      wp->flags |= WEBS_AUTH_DIGEST;
      break;

     case AuthBasic:
      //wp->flags &= ~WEBS_AUTH_DIGEST;
      break;
    }

  // ask for password if none provided
  if (!password)
    {
      websError(wp,401,passwordRequiredString);
      goto loginFail;
    }
  
  // we have a password, so check it
  if (ident.authMethod == AuthBasic)
    {
      if (strcmp(password,ident.password))
        {
          websError(wp,401,passwordRequiredString);
          goto loginFail;
        }
    }
  else if (flags & WEBS_AUTH_DIGEST)
    {
      char *digestCalc;
      wp->password = ident.password;
      a_assert(wp->digest);
      a_assert(wp->nonce);
      a_assert(wp->password);
      digestCalc = websCalcDigest(wp);
      a_assert(digestCalc);
      if (strcmp(wp->digest, digestCalc))
        {
          websError(wp,401,passwordRequiredString);
          goto loginFail;
        }
    }
  
 loginOK:
    {
      // send login e-mail if it's been more than 5 minutes
      // since last request -or- if last request was from
      // different IP address
      static long lastRequestTime;
      static char lastRequestIP[20];
      long now = time(NULL);
      
      if ((now - lastRequestTime) > (long)(60*5))
        *lastRequestIP = '\0';
      if (strcmp(wp->ipaddr,lastRequestIP))
        {
          //diag_printf("Web access OK from %s\n",wp->ipaddr);
          strcpy(lastRequestIP,wp->ipaddr);
        }
      lastRequestTime = now;
      return 0;
    }
  
 loginFail:
    {
      // send login failure message if wrong password was
      // supplied
      //diag_printf("Web access denied from %s\n",wp->ipaddr);
      if (password)
        {
          diag_printf("Wrong password\n");
        }
      return 1;
    }
}


//======================================================================
static int websHomePageHandler(webs_t wp, char_t * urlPrefix, char_t * webDir,
			       int arg, char_t * url, char_t * path,
			       char_t * query)
{
  if (*url == '\0' || gstrcmp(url,"/") == 0)
    {
      websRedirect(wp, T("home.asp"));
      return 1;
    }
  return 0;
}

//======================================================================
void defaultErrorHandler(int etype, char_t * msg)
{
  diag_printf(msg);
}

void defaultTraceHandler(int level, char_t * buf)
{
  if (buf && level <= GOAHEAD_TRACE_LEVEL)
    diag_printf("%s", buf);
}

#if !defined(CYGPKG_IO_FILEIO)
ssize_t send(int s, const void *buf, size_t len, int flags)
{
    return write(s, (void*)buf, len);
}

ssize_t recv(int s, void *buf, size_t len, int flags)
{
    return read(s, buf, len);
}
#endif

//======================================================================
#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_tables.h>
#include "../dm.h"


static cyg_handle_t webServerHandle;
static cyg_thread   webServerThread;
static unsigned webServerStack[2048];


/*
 *  Init function that starts the web server.
 */
void webServerInit(void)
{
  cyg_thread_create(20, webServerTask,
                    0,
                    "webs",
                    webServerStack,
                    sizeof webServerStack,
                    &webServerHandle,
                    &webServerThread);
  
  cyg_thread_resume(webServerHandle);
}

